part of fraction;

class CurrencyAmount extends Fraction {
  late Currency currency;
  late BigInt decimalScale;

  /// Returns a new currency amount instance from the unitless amount of token, i.e. the raw amount
  /// @param currency the currency in the amount
  /// @param rawAmount the raw token or ether amount
  factory CurrencyAmount.fromRawAmount(Currency currency, BigInt rawAmount) {
    assert(rawAmount <= MaxUint256, 'can max big Unit');
    return CurrencyAmount(currency, rawAmount);
  }

  /// Construct a currency amount with a denominator that is not equal to 1
  /// @param currency the currency
  /// @param numerator the numerator of the fractional token amount
  /// @param denominator the denominator of the fractional token amount
  factory CurrencyAmount.fromFractionalAmount(
    Currency currency,
    BigInt numerator,
    BigInt denominator,
  ) {
    return CurrencyAmount(currency, numerator, denominator: denominator);
  }

  /// Helper that calls the constructor with the ETHER currency
  /// @param amount ether amount in wei
  static CurrencyAmount ether(BigInt amount) {
    return CurrencyAmount(ETHER, amount);
  }

  BigInt get raw {
    return numerator;
  }

  CurrencyAmount(this.currency, BigInt numerator, {BigInt? denominator})
      : super(numerator, denominator: denominator ?? BigInt.one) {
    assert(super.quotient() <= MaxUint256, 'can max big AMOUNT');
    decimalScale = BigInt.from(10).pow(currency.decimals);
  }

  CurrencyAmount operator +(covariant CurrencyAmount other) {
    final added =  super.add(other);
    return CurrencyAmount.fromFractionalAmount(
        currency, added.numerator, added.denominator);
  }

  CurrencyAmount operator -(covariant CurrencyAmount other) {
    final subtracted = super.subtract(other);
    return CurrencyAmount.fromFractionalAmount(
        currency, subtracted.numerator, subtracted.denominator);
  }

  CurrencyAmount operator *(covariant CurrencyAmount other) {
    final multiply = super.multiply(other);
    return CurrencyAmount.fromFractionalAmount(
        currency, multiply.numerator, multiply.denominator);
  }

  CurrencyAmount operator /(covariant CurrencyAmount other) {
    final divide = super.divide(other);
    return CurrencyAmount.fromFractionalAmount(
        currency, divide.numerator, divide.denominator);
  }

  String toSignificant({
    int significantDigits = 6,
    NumberFormat? format,
    RoundingMode rounding = RoundingMode.roundDown,
  }) {
    final fraction = super.tryParseFraction(decimalScale);

    return super
        .divide(fraction)
        .toSignificantNative(significantDigits, format: format, rounding: rounding);
  }

  String toFixed({
    int? decimalPlaces,
    NumberFormat? format,
    RoundingMode rounding = RoundingMode.roundDown,
  }) {
    final decimal = decimalPlaces ?? currency.decimals;
    assert(decimal <= currency.decimals, 'DECIMALS');
    final fraction = super.tryParseFraction(decimalScale);

    return super
        .divide(fraction)
        .toFixedNative(decimal, format: format, rounding: rounding);
  }

  String toExact({
    NumberFormat? format,
  }) {
    Big.dp = currency.decimals;

    final valueDiv = super.quotient() / decimalScale;
    return Decimal.parse(valueDiv.toString()).toString();
  }

  @override
  bool operator ==(covariant CurrencyAmount other) {
    return currency == other.currency && currency.decimals == other.currency.decimals;
  }

  @override
  int get hashCode => currency.hashCode ^ currency.decimals.hashCode;
}
