part of fraction;

class Price extends Fraction {
  late Currency baseCurrency; // input i.e. denominator
  late Currency quoteCurrency; // output i.e. numerator
  late Fraction scalar; // used to adjust the raw fraction w/r/t the decimals of the {base,quote}Token

  /// Construct a price, either with the base and quote currency amount, or the
  /// @param args
  Price._(this.baseCurrency, this.quoteCurrency, this.scalar, BigInt numerator,
      BigInt? denominator)
      : super(numerator, denominator: denominator ?? BigInt.one);

  static Price fromRoute(Route route) {
    List<Price> prices = [];
    for (int i = 0; i < route.pairs.length; i++) {
      prices.add(route.path[i] == route.pairs[i].token0
          ? Price.fromContractor(
              route.pairs[i].reserve0.currency,
              route.pairs[i].reserve1.currency,
              route.pairs[i].reserve0.raw,
              route.pairs[i].reserve1.raw)
          : Price.fromContractor(
              route.pairs[i].reserve1.currency,
              route.pairs[i].reserve0.currency,
              route.pairs[i].reserve1.raw,
              route.pairs[i].reserve0.raw));
    }

    Price itemPrice = prices[0];

    if(prices.length>1){
      prices = prices.sublist(1);
      for (final ele in prices) {
        itemPrice = itemPrice * ele;
      }
    }

    return itemPrice;
  }

  factory Price.fromContractor(Currency baseCurrency, Currency quoteCurrency,
      BigInt denominator, BigInt numerator) {
    final scalar = Fraction(BigInt.from(pow(10, baseCurrency.decimals)),
        denominator: BigInt.from(pow(10, quoteCurrency.decimals)));

    return Price._(baseCurrency, quoteCurrency, scalar, numerator, denominator);
  }

  /// Flip the price, switching the base and quote currency
  @override
  Price invert() {
    return Price.fromContractor(
        quoteCurrency, baseCurrency, numerator, denominator);
  }

  /// Multiply the price by another price, returning a new price. The other price must have the same base currency as this price's quote currency
  Price operator *(covariant Price other) {
    assert(quoteCurrency == other.baseCurrency, 'TOKEN');
    final fraction = super.multiply(other);
    return Price.fromContractor(baseCurrency, other.quoteCurrency,
        fraction.denominator, fraction.numerator);
  }

  /// Return the amount of quote currency corresponding to a given amount of the base currency
  /// @param currencyAmount the amount of base currency to quote against the price
  CurrencyAmount quote(CurrencyAmount currencyAmount) {
    assert(currencyAmount.currency == baseCurrency, 'TOKEN');
    final result = super.multiply(currencyAmount);
    return CurrencyAmount.fromFractionalAmount(
        quoteCurrency, result.numerator, result.denominator);
  }

  /// Get the value scaled by decimals for formatting
  /// @private
  Fraction _adjustedForDecimals() {
    return super.multiply(scalar);
  }

  String toSignificant(
      {int significantDigits = 6,
      NumberFormat? format,
      RoundingMode? rounding}) {
    return _adjustedForDecimals()
        .toSignificantNative(significantDigits, format: format, rounding: rounding);
  }

  String toFixed(
      {int decimalPlaces = 6, NumberFormat? format, RoundingMode? rounding}) {
    return super.toFixedNative(decimalPlaces, format: format, rounding: rounding);
  }

  @override
  bool operator == (Object other) {
    if (other is Price) {
      // print("quoteCurrency: $baseCurrency == ${other.baseCurrency} ");
      // print("quoteCurrency: $quoteCurrency == ${other.quoteCurrency} ");
      // print("quoteCurrency: ${raw.denominator} == ${other.raw.denominator} ");
      // print("quoteCurrency: ${raw.numerator} == ${other.raw.numerator} ");
      return baseCurrency == other.baseCurrency && quoteCurrency == other.quoteCurrency && raw == other.raw;
    } else {
      return false;
    }
  }

  Fraction get raw {
    return Fraction(numerator, denominator: denominator);
  }

  @override
  int get hashCode => baseCurrency.hashCode ^ quoteCurrency.hashCode ^ raw.hashCode;
}
