import 'package:pancake_sdk/core/entities/fraction/index.dart';
import 'package:pancake_sdk/core/entities/token/tokenAmount.dart';
import 'package:pancake_sdk/core/entities/index.dart';
import 'package:pancake_sdk/core/extensions/bigint_ext.dart';
import 'package:pancake_sdk/core/utils/index.dart';
import 'package:pancake_sdk/v2/constants.dart';
import 'package:pancake_sdk/v2/errors.dart';
import 'package:pancake_sdk/v2/utils.dart';
import 'package:web3dart/web3dart.dart';

final PAIR_ADDRESS_CACHE = <String, EthereumAddress>{};

composeKey(Token token0, Token token1) =>
    '${token0.chainId}-${token0.address}-${token1.address}';

class Pair {
  static EthereumAddress? getAddress(Token tokenA, Token tokenB) {
    final pairTokens =
        tokenA.sortsBefore(tokenB) ? [tokenA, tokenB] : [tokenB, tokenA];
    final key = composeKey(pairTokens[0], pairTokens[1]);

    if (PAIR_ADDRESS_CACHE.containsKey(key) == false) {
      PAIR_ADDRESS_CACHE[key] = getCreate2Address(
        FACTORY_ADDRESS_MAP[pairTokens[0].chainId]!,
        '${pairTokens[0].address.hexNo0x}${pairTokens[1].address.hexNo0x}',
        INIT_CODE_HASH_MAP[pairTokens[0].chainId]!,
      );
    }

    return PAIR_ADDRESS_CACHE[key];
  }

  Token? liquidityToken;

  late List<TokenAmount> _tokenAmounts;
  late BigInt tokenAmounts0;
  late BigInt tokenAmounts1;

  Pair(TokenAmount tokenAmountA, TokenAmount tokenAmountB, {Token? liqToken}) {
    _tokenAmounts = tokenAmountA.token.sortsBefore(tokenAmountB.token)
        ? [tokenAmountA, tokenAmountB]
        : [tokenAmountB, tokenAmountA];
    tokenAmounts0 = _tokenAmounts[0].raw;
    tokenAmounts1 = _tokenAmounts[1].raw;

    liquidityToken ??= liqToken ??
        Token(
          _tokenAmounts.first.token.chainId,
          Pair.getAddress(_tokenAmounts[0].token, _tokenAmounts[1].token)!,
          18,
          symbol: 'Cake-LP',
          name: 'Pancake LPs',
        );
  }

  Price get token0Price => Price.fromContractor(
      token0, token1, _tokenAmounts[0].raw, _tokenAmounts[1].raw);

  Price get token1Price => Price.fromContractor(
      token1, token0, _tokenAmounts[1].raw, _tokenAmounts[0].raw);

  Price priceOf(Token token) {
    invariant(involvesToken(token), ' involves token $token');
    return token == token0 ? token0Price : token1Price;
  }

  int get chainId {
    return token0.chainId;
  }

  Token get token0 => _tokenAmounts[0].token;

  Token get token1 => _tokenAmounts[1].token;

  TokenAmount get reserve0 => _tokenAmounts[0];

  TokenAmount get reserve1 => _tokenAmounts[1];

  bool involvesToken(Token token) => token == token0 || token == token1;

  TokenAmount reserveOf(Token token) {
    invariant(involvesToken(token), ' involves token $token');
    return token == token0 ? reserve0 : reserve1;
  }

  /// returns list of values :
  /// 0 = outputAmount : [TokenAmount]
  /// 1 = nextPair : [Pair]
  List<dynamic> getOutputAmount(TokenAmount inputAmount) {
    invariant(involvesToken(inputAmount.token),
        ' involves token ${inputAmount.token}');

    if (reserve0.raw == BigInt.zero || reserve1.raw == BigInt.zero) {
      throw InsufficientReservesError();
    }
    var inputReserve = reserveOf(inputAmount.token);
    var outputToken = inputAmount.token == token0 ? token1 : token0;
    var outputReserve = reserveOf(outputToken);

    var inputAmountWithFee = inputAmount.raw * BI997;

    var outputAmount = TokenAmount(
      outputToken,
      (inputAmountWithFee * outputReserve.raw) ~/
          (inputReserve.raw * BI1000 + inputAmountWithFee),
    );

    if (outputAmount.raw == BigInt.zero) {
      throw InsufficientInputAmountError('output not zero');
    }
    return [
      outputAmount,
      Pair(inputReserve + inputAmount, outputReserve - outputAmount)
    ];
  }

  /// returns list of values :
  /// 0 = inputAmount : [TokenAmount]
  /// 1 = nextPair : [Pair]
  List<dynamic> getInputAmount(TokenAmount outputAmount) {
    invariant(involvesToken(outputAmount.token),
        ' involves token ${outputAmount.token}');

    if (reserve0.raw == BigInt.zero ||
        reserve1.raw == BigInt.zero ||
        outputAmount.raw >= reserveOf(outputAmount.token).raw) {
      throw InsufficientReservesError();
    }
    var outputReserve = reserveOf(outputAmount.token);
    var inputToken = outputAmount.token == token0 ? token1 : token0;
    var inputReserve = reserveOf(inputToken);
    var numerator = inputReserve.raw * outputAmount.raw * BI1000;
    var denominator = (outputReserve.raw - outputAmount.raw) * BI997;

    var inputAmount = TokenAmount(
      inputToken,
      (numerator ~/ denominator) + BigInt.one,
    );

    return [
      inputAmount,
      Pair(inputReserve + inputAmount, outputReserve - outputAmount)
    ];
  }

  TokenAmount getLiquidityMinted(
    TokenAmount totalSupply,
    TokenAmount tokenAmountA,
    TokenAmount tokenAmountB,
  ) {
    invariant(liquidityToken != null, 'liquidityToken is null');
    invariant(totalSupply.token == liquidityToken,
        'totalSupply token is not liquidity token');
    var tokenAmounts = tokenAmountA.token.sortsBefore(tokenAmountB.token)
        ? [tokenAmountA, tokenAmountB]
        : [tokenAmountB, tokenAmountA];

    invariant(totalSupply.token == liquidityToken,
        'totalSupply token is not liquidity token');
    invariant(
        tokenAmounts[0].token == token0 && tokenAmounts[1].token == token1,
        'tokenAmountA can\'t be like  tokenAmountB');

    BigInt liquidity;

    if (totalSupply.raw == BigInt.zero) {
      liquidity = (tokenAmounts[0].raw * tokenAmounts[1].raw).babylonianSqrt() -
          MINIMUM_LIQUIDITY;
    } else {
      var amount0 = (tokenAmounts[0].raw * totalSupply.raw) ~/ reserve0.raw;
      var amount1 = (tokenAmounts[1].raw * totalSupply.raw) ~/ reserve1.raw;

      liquidity = amount0 >= amount1 ? amount0 : amount1;
    }
    if (!(liquidity > BigInt.zero)) {
      throw InsufficientInputAmountError();
    }
    return TokenAmount(liquidityToken!, liquidity);
  }

  TokenAmount getLiquidityValue(
    Token token,
    TokenAmount totalSupply,
    TokenAmount liquidity,
    BigInt kLast, [
    bool feeOn = false,
  ]) {
    invariant(liquidityToken != null, 'liquidityToken is null');
    invariant(involvesToken(token), ' involves token $token');
    invariant(totalSupply.token == liquidityToken,
        'totalSupply token is not liquidity Token ');
    invariant(liquidity.token == liquidityToken, 'liquidity token incorrect');
    invariant(liquidity.raw <= totalSupply.raw,
        'InWei of liquidity token < totalSupply token');

    TokenAmount totalSupplyAdjusted;

    if (!feeOn) {
      totalSupplyAdjusted = totalSupply;
    } else {
      if (kLast != BigInt.zero) {
        var rootK = (reserve0.raw * reserve1.raw).babylonianSqrt();
        var rootKLast = kLast.babylonianSqrt();
        if (rootK > rootKLast) {
          final numerator = totalSupply.raw * (rootK - rootKLast);
          final denominator = rootK * BigInt.from(5) + rootKLast;
          final feeLiquidity = numerator ~/ denominator;
          totalSupplyAdjusted =
              totalSupply + TokenAmount(liquidityToken!, feeLiquidity);
        } else {
          totalSupplyAdjusted = totalSupply;
        }
      } else {
        totalSupplyAdjusted = totalSupply;
      }
    }

    return TokenAmount(
      token,
      (liquidity.raw * reserveOf(token).raw) ~/ totalSupplyAdjusted.raw,
    );
  }

  @override
  bool operator ==(Object other) {
    if (other is Pair) {
      return token0.address == other.token0.address &&
          token1.address == other.token1.address &&
          chainId == other.chainId;
    } else {
      return false;
    }
  }

  @override
  int get hashCode => token0.hashCode ^ token1.hashCode;
}

extension ExtendedIterable<E> on Iterable<E> {
  void forEachIndexed(void Function(E e, int i) f) {
    var i = 0;
    forEach((e) => f(e, i++));
  }
}
