import 'package:pancake_sdk/core/entities/fraction/index.dart';
import 'package:pancake_sdk/core/entities/index.dart';
import 'package:pancake_sdk/core/entities/token/tokenAmount.dart';
import 'package:pancake_sdk/core/utils/index.dart';
import 'package:pancake_sdk/v2/constants.dart';
import 'package:pancake_sdk/v2/errors.dart';
import 'package:pancake_sdk/v2/pair.dart';
import 'package:pancake_sdk/v2/route.dart';
import 'package:pancake_sdk/v2/utils.dart';

/// Returns the percent difference between the mid price and the execution price, i.e. price impact.
/// @param midPrice mid price before the trade
/// @param inputAmount the input amount of the trade
/// @param outputAmount the output amount of the trade
Percent computePriceImpact(
    Price midPrice, CurrencyAmount inputAmount, CurrencyAmount outputAmount) {
  final inputAmountFraction = Fraction(inputAmount.raw);
  final outputAmountFraction = Fraction(outputAmount.raw);
  var exactQuote = midPrice.raw.multiply(inputAmountFraction);
  var slippage = (exactQuote.subtract(outputAmountFraction).divide(exactQuote));
  return Percent(slippage.numerator, denominator: slippage.denominator);
}

/// minimal interface so the input output comparator may be shared across types
class InputOutput {
  CurrencyAmount inputAmount;
  CurrencyAmount outputAmount;

  InputOutput(this.inputAmount, this.outputAmount);
}

class BestTradeOptions {
// how many results to return
  double? maxNumResults;

// the maximum number of hops a trade should contain
  double? maxHops;

  BestTradeOptions({this.maxNumResults, this.maxHops});
}

// extension of the input output comparator that also considers other dimensions of the trade in ranking them
int tradeComparator(Trade a, Trade b) {
  var ioComp = inputOutputComparator(InputOutput(a.inputAmount, a.outputAmount),
      InputOutput(b.inputAmount, b.outputAmount));

  if (ioComp != 0) {
    return ioComp;
  }

  // consider lowest slippage next, since these are less likely to fail
  if (a.priceImpact.lessThan(b.priceImpact)) {
    return -1;
  } else if (a.priceImpact.greaterThan(b.priceImpact)) {
    return 1;
  }

  // finally consider the number of hops since each hop costs gas
  return a.route.path.length - b.route.pairs.length;
}

/// Given a currency amount and a chain ID, returns the equivalent representation as the token amount.
/// In other words, if the currency is ETHER, returns the WETH token amount for the given chain. Otherwise, returns
/// the input currency amount.
TokenAmount wrappedAmount(CurrencyAmount currencyAmount, ChainId chainId) {
  if (currencyAmount is TokenAmount) return currencyAmount;
  if(currencyAmount.currency is Token){
    if (currencyAmount.currency == Token.WETH[ChainId.MAINNET.value]) {
      return TokenAmount(Token.WETH[chainId.value]!, currencyAmount.raw);
    }
  }else{
    if (currencyAmount.currency == ETHER) {
      return TokenAmount(Token.WETH[chainId.value]!, currencyAmount.raw);
    }
  }
  if ((currencyAmount.currency.symbol ?? '').contains(ETHER.symbol ?? '')) {
    return TokenAmount(Token.WETH[chainId.value]!, currencyAmount.raw);
  }
  invariant(false, 'CURRENCY');
  return TokenAmount(Token.WETH[chainId.value]!, currencyAmount.raw);
}

Token? wrappedCurrency(Currency currency, int chainId) {
  if (currency is Token) return currency;
  if(currency is Token){
    if(currency == Token.WETH[ChainId.MAINNET.value]){
      return Token.WETH[chainId]!;
    }
  }else {
    if(currency == ETHER){
      return Token.WETH[chainId]!;
    }
  }
  return null;
}

/// comparator function that allows sorting trades by their output amounts, in decreasing order, and then input amounts
/// in increasing order. i.e. the best trades have the most outputs for the least inputs and are sorted first
int inputOutputComparator(InputOutput a, InputOutput b) {
  assert(a.inputAmount.currency == b.inputAmount.currency);
  assert(a.outputAmount.currency == b.outputAmount.currency);

  if (a.outputAmount == b.outputAmount) {
    if (a.inputAmount == b.inputAmount) {
      return 0;
    }
    // trade A requires less input than trade B, so A should come first
    if (a.inputAmount.raw < b.inputAmount.raw) {
      return -1;
    } else {
      return 1;
    }
  } else {
    // tradeA has less output than trade B, so should come second
    if (a.outputAmount.raw < b.outputAmount.raw) {
      return 1;
    } else {
      return -1;
    }
  }
}

/// Represents a trade executed against a list of pairs.
/// Does not account for slippage, i.e. trades that front run this trade and move the price.
class Trade {
  /// The route of the trade, i.e. which pairs the trade goes through.
  late Route route;

  /// The type of the trade, either exact in or exact out.
  late TradeType tradeType;

  /// The input amount for the trade assuming no slippage.
  late CurrencyAmount inputAmount;

  /// The output amount for the trade assuming no slippage.
  late CurrencyAmount outputAmount;

  /// The price expressed in terms of output amount/input amount.
  late Price executionPrice;

  /// The mid price after the trade executes assuming no slippage.
  late Price nextMidPrice;

  /// The percent difference between the mid price before the trade and the trade execution price.
  late Percent priceImpact;

  /// Constructs an exact in trade with the given amount in and route
  /// @param route route of the exact in trade
  /// @param amountIn the amount being passed in
  static Trade exactIn(Route route, CurrencyAmount amountIn) {
    return Trade(route, amountIn, TradeType.EXACT_INPUT);
  }

  /// Constructs an exact out trade with the given amount out and route
  /// @param route route of the exact out trade
  /// @param amountOut the amount returned by the trade
  static Trade exactOut(Route route, CurrencyAmount amountOut) {
    return Trade(route, amountOut, TradeType.EXACT_OUTPUT);
  }

  Trade(this.route, CurrencyAmount amount, this.tradeType) {
    var amounts =
        List<TokenAmount?>.generate(route.path.length, (index) => null);
    var nextPairs = List<Pair?>.generate(route.pairs.length, (index) => null);
    if (tradeType == TradeType.EXACT_INPUT) {
      assert(amount.currency == route.input);

      amounts[0] = wrappedAmount(amount, route.chainId);
      for (var i = 0; i < route.path.length - 1; i++) {
        var pair = route.pairs[i];
        var outputAmountAndNextPair = pair.getOutputAmount(amounts[i]!);
        amounts[i + 1] = outputAmountAndNextPair[0];
        nextPairs[i] = outputAmountAndNextPair[1];
      }
    } else {
      assert(amount.currency == route.output);

      amounts[amounts.length - 1] = wrappedAmount(amount, route.chainId);
      for (var i = route.path.length - 1; i > 0; i--) {
        var pair = route.pairs[i - 1];

        var inputAmountAndNextPair = pair.getInputAmount(amounts[i]!);
        amounts[i - 1] = inputAmountAndNextPair[0];
        nextPairs[i - 1] = inputAmountAndNextPair[1];
      }
    }

    inputAmount = (tradeType == TradeType.EXACT_INPUT
        ? amount
        : route.input == ETHER
            ? CurrencyAmount.ether(amounts.first!.raw)
            : amounts.first)!;

    outputAmount = (tradeType == TradeType.EXACT_OUTPUT
        ? amount
        : route.output == ETHER
            ? CurrencyAmount.ether(amounts[amounts.length - 1]!.raw)
            : amounts[amounts.length - 1]!);

    executionPrice = Price.fromContractor(inputAmount.currency,
        outputAmount.currency, inputAmount.raw, outputAmount.raw);
    priceImpact = computePriceImpact(route.midPrice, inputAmount, outputAmount);
  }

  /// Get the minimum amount that must be received from this trade for the given slippage tolerance
  /// @param slippageTolerance tolerance of unfavorable slippage from the execution price of this trade
  CurrencyAmount minimumAmountOut(Percent slippageTolerance) {
    invariant(!(slippageTolerance.denominator < ZERO), 'SLIPPAGE_TOLERANCE');
    invariant(!(slippageTolerance.numerator < ZERO), 'SLIPPAGE_TOLERANCE');
    final outputAmountFraction = Fraction(outputAmount.raw);
    if (tradeType == TradeType.EXACT_OUTPUT) {
      return outputAmount;
    } else {
      final slippageAdjustedAmountOut = Fraction(ONE)
          .add(slippageTolerance)
          .invert()
          .multiply(outputAmountFraction)
          .quotient();
      return outputAmount is TokenAmount
          ? TokenAmount(
              (outputAmount as TokenAmount).token, slippageAdjustedAmountOut)
          : CurrencyAmount.ether(
              slippageAdjustedAmountOut,
            );
    }
  }

  /// Get the maximum amount in that can be spent via this trade for the given slippage tolerance
  /// @param slippageTolerance tolerance of unfavorable slippage from the execution price of this trade
  CurrencyAmount maximumAmountIn(Percent slippageTolerance) {
    invariant(!(slippageTolerance.denominator < ZERO), 'SLIPPAGE_TOLERANCE');
    invariant(!(slippageTolerance.numerator < ZERO), 'SLIPPAGE_TOLERANCE');
    final inputAmountOutFraction = Fraction(inputAmount.raw);
    if (tradeType == TradeType.EXACT_INPUT) {
      return inputAmount;
    } else {
      final slippageAdjustedAmountIn = Fraction(ONE)
          .add(slippageTolerance)
          .multiply(inputAmountOutFraction)
          .quotient();
      return inputAmount is TokenAmount
          ? TokenAmount(
              (inputAmount as TokenAmount).token, slippageAdjustedAmountIn)
          : CurrencyAmount.ether(
              slippageAdjustedAmountIn,
            );
    }
  }

  /// Given a list of pairs, and a fixed amount in, returns the top `maxNumResults` trades that go from an input token
  /// amount to an output token, making at most `maxHops` hops.
  /// Note this does not consider aggregation, as routes are linear. It's possible a better route exists by splitting
  /// the amount in among multiple routes.
  /// @param pairs the pairs to consider in finding the best trade
  /// @param currencyAmountIn exact amount of input currency to spend
  /// @param currencyOut the desired currency out
  /// @param maxNumResults maximum number of results to return
  /// @param maxHops maximum number of hops a returned trade can make, e.g. 1 hop goes through a single pair
  /// @param currentPairs used in recursion; the current list of pairs
  /// @param originalAmountIn used in recursion; the original value of the currencyAmountIn parameter
  /// @param bestTrades used in recursion; the current list of best trades
  static List<Trade> bestTradeExactIn(
    List<Pair> pairs,
    CurrencyAmount currencyAmountIn,
    Currency currencyOut, {
    int maxNumResults = 3,
    int maxHops = 3,
    // used in recursion.
    List<Pair>? currentPairs,
    CurrencyAmount? originalAmountIn,
    List<Trade>? bestTrades,
  }) {
    currentPairs = currentPairs ?? [];
    originalAmountIn = originalAmountIn ?? currencyAmountIn;
    bestTrades = bestTrades ?? [];
    invariant(pairs.isNotEmpty, 'PAIRS');
    invariant(maxHops > 0, 'MAX_HOPS');
    invariant(originalAmountIn == currencyAmountIn || currentPairs.isNotEmpty,
        'INVALID_RECURSION');
    final chainId = currencyAmountIn is TokenAmount
        ? currencyAmountIn.token.chainId
        : currencyOut is Token
            ? currencyOut.chainId
            : null;
    invariant(chainId != null, 'CHAIN_ID');

    final amountIn =
        wrappedAmount(currencyAmountIn, int.parse(chainId.toString()).chain);
    final tokenOut =
        wrappedCurrency(currencyOut, int.parse(chainId.toString()));
    for (int i = 0; i < pairs.length; i++) {
      final pair = pairs[i];
      // pair irrelevant
      if (pair.token0 != amountIn.token && pair.token1 != amountIn.token) {
        continue;
      }
      if (pair.reserve0.raw == ZERO || pair.reserve1.raw == ZERO) continue;

      TokenAmount amountOut;
      try {
        amountOut = pair.getOutputAmount(amountIn)[0];
      } on InsufficientReservesError catch (_) {
        continue;
      } catch (error) {
        continue;
      }
      // we have arrived at the output token, so this is the final trade of one of the paths
      if (amountOut.token == tokenOut) {
        sortedInsertTrade(
          bestTrades,
          Trade(
              Route([...currentPairs, pair], originalAmountIn.currency,
                  output: currencyOut),
              originalAmountIn,
              TradeType.EXACT_INPUT),
          maxNumResults,
        );
      } else if (maxHops > 1 && pairs.length > 1) {
        final dataLastPair = pairs.sublist(i + 1, pairs.length);
        final pairsExcludingThisPair = pairs.sublist(0, i) + dataLastPair;

        // otherwise, consider all the other paths that lead from this token as long as we have not exceeded maxHops
        Trade.bestTradeExactIn(pairsExcludingThisPair, amountOut, currencyOut,
            maxNumResults: maxNumResults,
            maxHops: (maxHops - 1),
            currentPairs: [...currentPairs, pair],
            originalAmountIn: originalAmountIn,
            bestTrades: bestTrades);
      }
    }

    return bestTrades;
  }

  /// similar to the above method but instead targets a fixed output amount
  /// given a list of pairs, and a fixed amount out, returns the top `maxNumResults` trades that go from an input token
  /// to an output token amount, making at most `maxHops` hops
  /// note this does not consider aggregation, as routes are linear. it's possible a better route exists by splitting
  /// the amount in among multiple routes.
  /// @param pairs the pairs to consider in finding the best trade
  /// @param currencyIn the currency to spend
  /// @param currencyAmountOut the exact amount of currency out
  /// @param maxNumResults maximum number of results to return
  /// @param maxHops maximum number of hops a returned trade can make, e.g. 1 hop goes through a single pair
  /// @param currentPairs used in recursion; the current list of pairs
  /// @param originalAmountOut used in recursion; the original value of the currencyAmountOut parameter
  /// @param bestTrades used in recursion; the current list of best trades
  static List<Trade> bestTradeExactOut(
    List<Pair> pairs,
    Currency currencyIn,
    CurrencyAmount currencyAmountOut, {
    int maxNumResults = 3,
    int maxHops = 3,
    List<Pair>? currentPairs,
    CurrencyAmount? originalAmountOut,
    List<Trade>? bestTrades,
  }) {
    originalAmountOut ??= currencyAmountOut;
    currentPairs ??= <Pair>[];
    bestTrades ??= <Trade>[];

    invariant(pairs.isNotEmpty, 'PAIRS');
    invariant(maxHops > 0, 'MAX_HOPS');
    invariant(originalAmountOut == currencyAmountOut || currentPairs.isNotEmpty,
        'INVALID_RECURSION');

    var chainId = currencyAmountOut is TokenAmount
        ? currencyAmountOut.token.chainId
        : currencyIn is Token
            ? currencyIn.chainId
            : null;

    invariant(chainId != null, 'CHAIN_ID');

    var amountOut = wrappedAmount(currencyAmountOut, chainId!.chain);
    var tokenIn = wrappedCurrency(currencyIn, chainId);

    for (var i = 0; i < pairs.length; i++) {
      final pair = pairs[i];

      // pair irrelevant
      if (pair.token0 != amountOut.token && pair.token1 != amountOut.token) {
        continue;
      }

      if (pair.reserve0.raw == BigInt.zero || pair.reserve1.raw == BigInt.zero) {
        continue;
      }

      late TokenAmount amountIn;
      try {
        amountIn = pair.getInputAmount(amountOut)[0];

        // input too low
      } on InsufficientReservesError catch (_) {
        continue;
      } catch (e) {
        rethrow;
      }

      // we have arrived at the output token, so this is the final trade of one of the paths
      if (amountIn.token == tokenIn) {
        sortedInsertTrade(
          bestTrades,
          Trade(
            Route([pair, ...currentPairs], currencyIn,
                output: originalAmountOut.currency),
            originalAmountOut,
            TradeType.EXACT_OUTPUT,
          ),
          maxNumResults,
        );
      } else if (maxHops > 1 && pairs.length > 1) {
        final dataLastPair = pairs.sublist(i + 1, pairs.length);
        final pairsExcludingThisPair = pairs.sublist(0, i) + dataLastPair;
        // otherwise, consider all the other paths that lead from this token as long as we have not exceeded maxHops
        Trade.bestTradeExactOut(
          pairsExcludingThisPair,
          currencyIn,
          amountIn,
          maxNumResults: maxNumResults,
          maxHops: maxHops - 1,
          currentPairs: [pair, ...currentPairs],
          originalAmountOut: originalAmountOut,
          bestTrades: bestTrades,
        );
      }
    }
    return bestTrades;
  }
}
