import 'package:pancake_sdk/core/entities/fraction/index.dart';
import 'package:pancake_sdk/core/entities/index.dart';
import 'package:pancake_sdk/core/entities/token/tokenAmount.dart';
import 'package:pancake_sdk/v2/constants.dart';
import 'package:pancake_sdk/v2/pair.dart';
import 'package:test/test.dart';
import 'package:web3dart/web3dart.dart';

void main() {
  group('Pair', () {
    final USDC = Token(ChainId.MAINNET.value, EthereumAddress.fromHex('0x8AC76a51cc950d9822D68b83fE1Ad97B32Cd580d'), 18, symbol: 'USDC', name: 'USD Coin');
    final DAI = Token(ChainId.MAINNET.value, EthereumAddress.fromHex('0x1AF3F329e8BE154074D8769D1FFa4eE058B1DBc3'), 18, symbol: 'DAI', name: 'DAI Stablecoin');

    group('#constructor', () {
      test('cannot be used for tokens on different chains', () => {
            expect(() => Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(Token.WETH[ChainId.TESTNET.value]!, BigInt.from(100))),
              throwsA(isA<AssertionError>()),
            )
          });
    });

    group('#getAddress', () {
      test('returns the correct address', () {
        expect(
            Pair.getAddress(USDC, DAI),
            equals(EthereumAddress.fromHex('0xadBba1EF326A33FDB754f14e62A96D5278b942Bd')));
      });
    });

    group('#token0', () {
      test('always is the token that sorts before', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).token0,
            DAI);

        expect(
            Pair(TokenAmount(DAI, BigInt.from(100)), TokenAmount(USDC, BigInt.from(100))).token0,
            DAI);
      });
    });

    group('#token1', () {
      test('always is the token that sorts after', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).token1,
            USDC);

        expect(
            Pair(TokenAmount(DAI, BigInt.from(100)), TokenAmount(USDC, BigInt.from(100))).token1,
            USDC);
      });
    });

    group('#reserve0', () {
      test('always comes from the token that sorts before', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(101))).reserve0,
            TokenAmount(DAI, BigInt.from(101)));

        expect(
            Pair(TokenAmount(DAI, BigInt.from(101)), TokenAmount(USDC, BigInt.from(100))).reserve0,
            TokenAmount(DAI, BigInt.from(101)));
      });
    });

    group('#reserve1', () {
      test('always comes from the token that sorts after', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(101))).reserve1,
            TokenAmount(USDC, BigInt.from(100)));

        expect(
            Pair(TokenAmount(DAI, BigInt.from(101)), TokenAmount(USDC, BigInt.from(100))).reserve1,
            TokenAmount(USDC, BigInt.from(100)));
      });
    });

    group('#token0Price', () {
      test('returns price of token0 in terms of token1', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(101)), TokenAmount(DAI, BigInt.from(100))).token0Price,
            Price.fromContractor(DAI,USDC,BigInt.from(100), BigInt.from(101)) );

        expect(
            Pair(TokenAmount(DAI, BigInt.from(100)), TokenAmount(USDC, BigInt.from(101))).token0Price,
            Price.fromContractor(DAI,USDC,BigInt.from(100), BigInt.from(101)));
      });
    });

    group('#token1Price', () {
      test('returns price of token0 in terms of token0', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(101)), TokenAmount(DAI, BigInt.from(100))).token1Price,
            Price.fromContractor(USDC,DAI,BigInt.from(101), BigInt.from(100)) );

        expect(
            Pair(TokenAmount(DAI, BigInt.from(100)), TokenAmount(USDC, BigInt.from(101))).token1Price,
            Price.fromContractor(USDC,DAI, BigInt.from(101), BigInt.from(100)));
      });
    });

    group('#priceOf()', () {
      final pair = Pair(
        TokenAmount(USDC, BigInt.from(101)),
        TokenAmount(DAI, BigInt.from(100)),
      );
      test('returns price of token in terms of other token', () {
        expect(pair.priceOf(DAI), pair.token0Price);
        expect(pair.priceOf(USDC), pair.token1Price);
      });

      test('throws if invalid token', () {
        expect(() => pair.priceOf(Token.WETH[ChainId.MAINNET.value]!), throwsA(isA<AssertionError>()));
      });
    });

    group('#reserveOf', () {
      test('returns reserves of the given token', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(101))).reserveOf(USDC),
            TokenAmount(USDC, BigInt.from(100)));

        expect(
            Pair(TokenAmount(DAI, BigInt.from(101)), TokenAmount(USDC, BigInt.from(100))).reserveOf(USDC),
            TokenAmount(USDC, BigInt.from(100)));
      });
    });

    group('#chainId', () {
      test('returns the token0 chainId', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).chainId,
            ChainId.MAINNET.value);

        expect(
            Pair(TokenAmount(DAI, BigInt.from(100)), TokenAmount(USDC, BigInt.from(100))).chainId,
            ChainId.MAINNET.value);
      });
    });

    group('#involvesToken', () {
      test('involves token', () {
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).involvesToken(USDC),
            true);

        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).involvesToken(DAI),
            true);
        expect(
            Pair(TokenAmount(USDC, BigInt.from(100)), TokenAmount(DAI, BigInt.from(100))).involvesToken(Token.WETH[ChainId.MAINNET.value]!),
            false);
      });
    });
  });
}
