import 'package:flutter_test/flutter_test.dart';
import 'package:pancake_sdk/core/entities/fraction/index.dart';
import 'package:pancake_sdk/core/entities/index.dart';
import 'package:pancake_sdk/core/entities/token/tokenAmount.dart';
import 'package:pancake_sdk/v2/constants.dart';
import 'package:pancake_sdk/v2/pair.dart';
import 'package:pancake_sdk/v2/route.dart';
import 'package:pancake_sdk/v2/router.dart';
import 'package:pancake_sdk/v2/trade.dart';
import 'package:web3dart/credentials.dart';

void main() {
  final token0 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000001'), 18,
      symbol: 't0');
  final token1 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000002'), 18,
      symbol: 't1');

  final pair_0_1 = Pair(TokenAmount(token0, BigInt.from(1000)),
      TokenAmount(token1, BigInt.from(1000)));

  final pairWeth0 = Pair(
      TokenAmount(Token.WETH[ChainId.MAINNET.value]!, BigInt.from(1000)),
      TokenAmount(token0, BigInt.from(1000)));

  void checkDeadline(String deadline) {
// less than 5 seconds on the deadline
    final dateTimeNow = DateTime.now();
    expect(dateTimeNow.millisecondsSinceEpoch / 1000 - int.parse(deadline) <= 5,
        true);
  }

  group('#swapCallParameters', () {
    group('exact in', () {
      test('ether to token1', () {
        final result = Router.swapCallParameters(
            Trade.exactIn(Route([pairWeth0, pair_0_1], ETHER, output: token1),
                CurrencyAmount.ether(BigInt.from(100))),
            TradeOptions(
                ttl: 50,
                recipient: '0x0000000000000000000000000000000000000004',
                allowedSlippage:
                    Percent(BigInt.from(1), denominator: BigInt.from(100))));

        List<String> listAddressTest = [
          Token.WETH[ChainId.MAINNET.value]!.address.toString(),
          token0.address.toString(),
          token1.address.toString(),
        ];

        expect(result.methodName, 'swapExactETHForTokens');
        expect(result.args.sublist(0, result.args.length - 1), [
          '0x51',
          listAddressTest,
          '0x0000000000000000000000000000000000000004'
        ]);
        expect(result.value, '0x64');
        checkDeadline(result.args[result.args.length - 1]);
      });
    });

    test('token1 to ether', () {
      final result = Router.swapCallParameters(
          Trade.exactIn(Route([pair_0_1, pairWeth0], token1, output: ETHER),
              TokenAmount(token1, BigInt.from(100))),
          TradeOptions(
              ttl: 50,
              recipient: '0x0000000000000000000000000000000000000004',
              allowedSlippage:
                  Percent(BigInt.from(1), denominator: BigInt.from(100))));
      expect(result.methodName, 'swapExactTokensForETH');
      expect(result.args.sublist(0, result.args.length - 1), [
        '0x64',
        '0x51',
        [
          token1.address.toString(),
          token0.address.toString(),
          Token.WETH[ChainId.MAINNET.value]!.address.toString()
        ],
        '0x0000000000000000000000000000000000000004'
      ]);
      expect(result.value, '0x0');
      checkDeadline(result.args[result.args.length - 1]);
    });

    test('token0 to token1', () {
      final result = Router.swapCallParameters(
          Trade.exactIn(Route([pair_0_1], token0, output: token1),
              TokenAmount(token0, BigInt.from(100))),
          TradeOptions(
              ttl: 50,
              recipient: '0x0000000000000000000000000000000000000004',
              allowedSlippage:
                  Percent(BigInt.from(1), denominator: BigInt.from(100))));
      expect(result.methodName, 'swapExactTokensForTokens');
      expect(result.args.sublist(0, result.args.length - 1), [
        '0x64',
        '0x59',
        [
          token0.address.toString(),
          token1.address.toString(),
        ],
        '0x0000000000000000000000000000000000000004'
      ]);
      expect(result.value, '0x0');
      checkDeadline(result.args[result.args.length - 1]);
    });
  });

  group('# exact out', () {
    test('ether to token1', () {
      final result = Router.swapCallParameters(
          Trade.exactOut(Route([pairWeth0, pair_0_1], ETHER, output: token1),
              TokenAmount(token1, BigInt.from(100))),
          TradeOptions(
              ttl: 50,
              recipient: '0x0000000000000000000000000000000000000004',
              allowedSlippage:
                  Percent(BigInt.from(1), denominator: BigInt.from(100))));
      expect(result.methodName, 'swapETHForExactTokens');
      expect(result.args.sublist(0, result.args.length - 1), [
        '0x64',
        [
          Token.WETH[ChainId.MAINNET.value]!.address.toString(),
          token0.address.toString(),
          token1.address.toString()
        ],
        '0x0000000000000000000000000000000000000004'
      ]);
      expect(result.value, '0x80');
      checkDeadline(result.args[result.args.length - 1]);
    });

    test('token1 to ether', () {
      final result = Router.swapCallParameters(
          Trade.exactOut(Route([pair_0_1, pairWeth0], token1, output: ETHER),
              CurrencyAmount.ether(BigInt.from(100))),
          TradeOptions(
              ttl: 50,
              recipient: '0x0000000000000000000000000000000000000004',
              allowedSlippage:
                  Percent(BigInt.from(1), denominator: BigInt.from(100))));
      expect(result.methodName, 'swapTokensForExactETH');
      expect(result.args.sublist(0, result.args.length - 1), [
        '0x64',
        '0x80',
        [
          token1.address.toString(),
          token0.address.toString(),
          Token.WETH[ChainId.MAINNET.value]!.address.toString()
        ],
        '0x0000000000000000000000000000000000000004'
      ]);
      expect(result.value, '0x0');
      checkDeadline(result.args[result.args.length - 1]);
    });

    test('token0 to token1', () {
      final result = Router.swapCallParameters(
          Trade.exactOut(Route([pair_0_1], token0, output: token1),
              TokenAmount(token1, BigInt.from(100))),
          TradeOptions(
              ttl: 50,
              recipient: '0x0000000000000000000000000000000000000004',
              allowedSlippage:
                  Percent(BigInt.from(1), denominator: BigInt.from(100))));
      expect(result.methodName, 'swapTokensForExactTokens');
      expect(result.args.sublist(0, result.args.length - 1), [
        '0x64',
        '0x71',
        [token0.address.toString(), token1.address.toString()],
        '0x0000000000000000000000000000000000000004'
      ]);
      expect(result.value, '0x0');
      checkDeadline(result.args[result.args.length - 1]);
    });
  });

  group('#supporting fee on transfer', () {
    group('#exact in', () {
      test('ether to token1', () {
        final result = Router.swapCallParameters(
            Trade.exactIn(Route([pairWeth0, pair_0_1], ETHER, output: token1),
                CurrencyAmount.ether(BigInt.from(100))),
            TradeOptions(
                ttl: 50,
                recipient: '0x0000000000000000000000000000000000000004',
                allowedSlippage:
                    Percent(BigInt.from(1), denominator: BigInt.from(100)),
                feeOnTransfer: true));
        expect(result.methodName,
            'swapExactETHForTokensSupportingFeeOnTransferTokens');
        expect(result.args.sublist(0, result.args.length - 1), [
          '0x51',
          [
            Token.WETH[ChainId.MAINNET.value]!.address.toString(),
            token0.address.toString(),
            token1.address.toString()
          ],
          '0x0000000000000000000000000000000000000004'
        ]);
        expect(result.value, '0x64');
        checkDeadline(result.args[result.args.length - 1]);
      });

      test('token1 to ether', () {
        final result = Router.swapCallParameters(
            Trade.exactIn(Route([pair_0_1, pairWeth0], token1, output: ETHER),
                TokenAmount(token1, BigInt.from(100))),
            TradeOptions(
                ttl: 50,
                recipient: '0x0000000000000000000000000000000000000004',
                allowedSlippage:
                    Percent(BigInt.from(1), denominator: BigInt.from(100)),
                feeOnTransfer: true));
        expect(result.methodName,
            'swapExactTokensForETHSupportingFeeOnTransferTokens');
        expect(result.args.sublist(0, result.args.length - 1), [
          '0x64',
          '0x51',
          [
            token1.address.toString(),
            token0.address.toString(),
            Token.WETH[ChainId.MAINNET.value]!.address.toString()
          ],
          '0x0000000000000000000000000000000000000004'
        ]);
        expect(result.value, '0x0');
        checkDeadline(result.args[result.args.length - 1]);
      });

      test('token0 to token1', () {
        final result = Router.swapCallParameters(
            Trade.exactIn(Route([pair_0_1], token0, output: token1),
                TokenAmount(token0, BigInt.from(100))),
            TradeOptions(
                ttl: 50,
                recipient: '0x0000000000000000000000000000000000000004',
                allowedSlippage:
                    Percent(BigInt.from(1), denominator: BigInt.from(100)),
                feeOnTransfer: true));
        expect(result.methodName,
            'swapExactTokensForTokensSupportingFeeOnTransferTokens');
        expect(result.args.sublist(0, result.args.length - 1), [
          '0x64',
          '0x59',
          [token0.address.toString(), token1.address.toString()],
          '0x0000000000000000000000000000000000000004'
        ]);
        expect(result.value, '0x0');
        checkDeadline(result.args[result.args.length - 1]);
      });
    });

    group('#exact out', () {
      test('ether to token1', () {
        expect(
          () => Router.swapCallParameters(
              Trade.exactOut(
                  Route([pairWeth0, pair_0_1], ETHER, output: token1),
                  TokenAmount(token1, BigInt.from(100))),
              TradeOptions(
                  ttl: 50,
                  recipient: '0x0000000000000000000000000000000000000004',
                  allowedSlippage:
                      Percent(BigInt.from(1), denominator: BigInt.from(100)),
                  feeOnTransfer: true)),
          throwsA(isA<AssertionError>()),
        );
      });

      test('token1 to ether', () {
        expect(
          () => Router.swapCallParameters(
              Trade.exactOut(
                  Route([pair_0_1, pairWeth0], token1, output: ETHER),
                  CurrencyAmount.ether(BigInt.from(100))),
              TradeOptions(
                  ttl: 50,
                  recipient: '0x0000000000000000000000000000000000000004',
                  allowedSlippage:
                      Percent(BigInt.from(1), denominator: BigInt.from(100)),
                  feeOnTransfer: true)),
          throwsA(isA<AssertionError>()),
        );
      });

      test('token0 to token1', () {
        expect(
          () => Router.swapCallParameters(
              Trade.exactOut(Route([pair_0_1], token0, output: token1),
                  TokenAmount(token1, BigInt.from(100))),
              TradeOptions(
                  ttl: 50,
                  recipient: '0x0000000000000000000000000000000000000004',
                  allowedSlippage:
                      Percent(BigInt.from(1), denominator: BigInt.from(100)),
                  feeOnTransfer: true)),
          throwsA(isA<AssertionError>()),
        );
      });
    });
  });
}
