import 'package:flutter_test/flutter_test.dart';
import 'package:pancake_sdk/core/entities/fraction/index.dart';
import 'package:pancake_sdk/core/entities/index.dart';
import 'package:pancake_sdk/core/entities/token/tokenAmount.dart';
import 'package:pancake_sdk/v2/constants.dart';
import 'package:pancake_sdk/v2/pair.dart';
import 'package:pancake_sdk/v2/route.dart';
import 'package:pancake_sdk/v2/trade.dart';
import 'package:web3dart/credentials.dart';

void main() {
  final token0 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000001'), 18,
      symbol: 't0');
  final token1 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000002'), 18,
      symbol: 't1');
  final token2 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000003'), 18,
      symbol: 't2');
  final token3 = Token(ChainId.MAINNET.value,
      EthereumAddress.fromHex('0x0000000000000000000000000000000000000005'), 18,
      symbol: 't3');

  final pair_0_1 = Pair(TokenAmount(token0, BigInt.from(1000)),
      TokenAmount(token1, BigInt.from(1000)));
  final pair_0_2 = Pair(TokenAmount(token0, BigInt.from(1000)),
      TokenAmount(token2, BigInt.from(1100)));
  final pair_0_3 = Pair(TokenAmount(token0, BigInt.from(1000)),
      TokenAmount(token3, BigInt.from(900)));
  final pair_1_2 = Pair(TokenAmount(token1, BigInt.from(1200)),
      TokenAmount(token2, BigInt.from(1000)));
  final pair_1_3 = Pair(TokenAmount(token1, BigInt.from(1200)),
      TokenAmount(token3, BigInt.from(1300)));

  final pairWeth0 = Pair(
      TokenAmount(Token.WETH[ChainId.MAINNET.value]!, BigInt.from(1000)),
      TokenAmount(token0, BigInt.from(1000)));

  final emptyPair01 = Pair(
      TokenAmount(token0, BigInt.from(0)), TokenAmount(token1, BigInt.from(0)));

  test('can be constructed with ETHER as input', () async {
    final trade = Trade(Route([pairWeth0], ETHER),
        CurrencyAmount.ether(BigInt.from(100)), TradeType.EXACT_INPUT);
    expect(trade.inputAmount.currency, ETHER);
    expect(trade.outputAmount.currency, token0);
  });

  test('can be constructed with ETHER as output', () async {
    final trade = Trade(Route([pairWeth0], token0, output: ETHER),
        TokenAmount(token0, BigInt.from(100)), TradeType.EXACT_INPUT);
    expect(trade.inputAmount.currency, token0);
    expect(trade.outputAmount.currency, ETHER);
  });

  group('#bestTradeExactIn', () {
    test(
        'throws with empty pairs',
            () =>
            expect(
                    () =>
                    Trade.bestTradeExactIn(
                        [], TokenAmount(token0, BigInt.from(100)), token2),
                throwsA(isA<AssertionError>())));

    test(
        'throws with max hops of 0',
            () =>
        {
          expect(
                  () =>
                  Trade.bestTradeExactIn(
                      [pair_0_2], TokenAmount(token0, BigInt.from(100)), token2,
                      maxHops: 0),
              throwsA(isA<AssertionError>()))
        });

    test('provides best route', () {
      final result = Trade.bestTradeExactIn([pair_0_1, pair_0_2, pair_1_2],
          TokenAmount(token0, BigInt.from(100)), token2);
      expect(result.length, 2);
      expect(result[0].route.pairs.length, 1); // 0 -> 2 at 10:11
      expect(result[0].route.path, [token0, token2]);
      expect(result[0].inputAmount, TokenAmount(token0, BigInt.from(100)));
      expect(result[0].outputAmount, TokenAmount(token2, BigInt.from(99)));
      expect(result[1].route.pairs.length, 2); // 0 -> 1 -> 2 at 12:12:10
      expect(result[1].route.path, [token0, token1, token2]);
      expect(result[1].inputAmount, TokenAmount(token0, BigInt.from(100)));
      expect(result[1].outputAmount, TokenAmount(token2, BigInt.from(69)));
    });

    test('doesnt throw for zero liquidity pairs', () {
      expect(
          Trade
              .bestTradeExactIn([emptyPair01],
              TokenAmount(token0, BigInt.from(100)), token1)
              .length,
          0);
    });

    test('respects maxHops', () {
      final result = Trade.bestTradeExactIn([
        pair_0_1,
        pair_0_2,
        pair_1_2
      ], TokenAmount(token0, BigInt.from(10)), token2, maxHops: 1);
      expect(result.length, 1);
      expect(result[0].route.pairs.length, 1); // 0 -> 2 at 10:11
      expect(result[0].route.path, [token0, token2]);
    });

    test('insufficient input for one pair', () {
      final result = Trade.bestTradeExactIn([pair_0_1, pair_0_2, pair_1_2],
          TokenAmount(token0, BigInt.from(1)), token2);
      expect(result.length, 1);
      expect(result[0].route.pairs.length, 1); // 0 -> 2 at 10:11
      expect(result[0].route.path, [token0, token2]);
      expect(result[0].outputAmount, TokenAmount(token2, BigInt.from(1)));
    });

    test('respects n', () {
      final result = Trade.bestTradeExactIn([
        pair_0_1,
        pair_0_2,
        pair_1_2
      ], TokenAmount(token0, BigInt.from(10)), token2, maxHops: 1);

      expect(result.length, 1);
    });

    test('no path', () {
      final result = Trade.bestTradeExactIn([pair_0_1, pair_0_3, pair_1_3],
          TokenAmount(token0, BigInt.from(10)), token2);
      expect(result.length, 0);
    });

    test('works for ETHER currency input', () {
      final result = Trade.bestTradeExactIn(
          [pairWeth0, pair_0_1, pair_0_3, pair_1_3],
          CurrencyAmount.ether(BigInt.from(100)),
          token3);
      expect(result.length, 2);
      expect(result[0].inputAmount.currency, ETHER);
      expect(result[0].route.path,
          [Token.WETH[ChainId.MAINNET.value], token0, token1, token3]);
      expect(result[0].outputAmount.currency, token3);
      expect(result[1].inputAmount.currency, ETHER);
      expect(
          result[1].route.path, [Token.WETH[ChainId.MAINNET.value], token0, token3]);
      expect(result[1].outputAmount.currency, token3);
    });

    test('works for ETHER currency output', () {
      final result = Trade.bestTradeExactIn(
          [pairWeth0, pair_0_1, pair_0_3, pair_1_3],
          TokenAmount(token3, BigInt.from(100)),
          ETHER);
      expect(result.length, 2);
      expect(result[0].inputAmount.currency, token3);
      expect(
          result[0].route.path, [token3, token0, Token.WETH[ChainId.MAINNET.value]]);
      expect(result[0].outputAmount.currency, ETHER);
      expect(result[1].inputAmount.currency, token3);
      expect(result[1].route.path,
          [token3, token1, token0, Token.WETH[ChainId.MAINNET.value]]);
      expect(result[1].outputAmount.currency, ETHER);
    });
  });

  group('#maximumAmountIn', () {
    group('#radeType = EXACT_INPUT', () {
      final exactIn = Trade(Route([pair_0_1, pair_1_2], token0),
          TokenAmount(token0, BigInt.from(100)), TradeType.EXACT_INPUT);
      test('throws if less than 0', () {
        expect(
                () =>
                exactIn.maximumAmountIn(
                    Percent(BigInt.from(-1), denominator: BigInt.from(100))),
            throwsA(isA<AssertionError>()));
      });

      test('returns exact if 0', () {
        expect(
            exactIn.maximumAmountIn(
                Percent(BigInt.from(0), denominator: BigInt.from(100))),
            exactIn.inputAmount);
      });

      test('returns exact if nonzero', () {
        expect(
            exactIn.maximumAmountIn(
                Percent(BigInt.from(0), denominator: BigInt.from(100))),
            TokenAmount(token0, BigInt.from(100)));
        expect(
            exactIn.maximumAmountIn(
                Percent(BigInt.from(5), denominator: BigInt.from(100))),
            TokenAmount(token0, BigInt.from(100)));
        expect(
            exactIn.maximumAmountIn(
                Percent(BigInt.from(200), denominator: BigInt.from(100))),
            TokenAmount(token0, BigInt.from(100)));
      });
    });
  });


  group('#minimumAmountOut', () {
    group('tradeType = EXACT_INPUT', () {
      final exactIn = Trade(
          Route([pair_0_1, pair_1_2], token0),
          TokenAmount(token0, BigInt.from(100)),
          TradeType.EXACT_INPUT
      );

      test('throws if less than 0', () {
        expect(() =>
            exactIn.minimumAmountOut(
                Percent(BigInt.from(-1), denominator: BigInt.from(100))),
            throwsA(isA<AssertionError>()));
      });

      test('returns exact if 0', () => {
      expect(exactIn.minimumAmountOut(Percent(BigInt.from(0), denominator: BigInt.from(100))), exactIn.outputAmount)
      });

      test('returns exact if nonzero', () {
        expect(exactIn.minimumAmountOut(Percent(BigInt.from(0),denominator: BigInt.from(100))),TokenAmount(token2, BigInt.from(69)));
        expect(exactIn.minimumAmountOut( Percent(BigInt.from(5),denominator: BigInt.from(100))),TokenAmount(token2,BigInt.from(65)));
        expect(exactIn.minimumAmountOut(Percent(BigInt.from(200),denominator: BigInt.from(100))),TokenAmount(token2, BigInt.from(23)));
      });
    });
  });

  group('tradeType = EXACT_OUTPUT', () {
    final exactOut = Trade(
        Route([pair_0_1, pair_1_2], token0),
        TokenAmount(token2, BigInt.from(100)),
        TradeType.EXACT_OUTPUT
    );

    test('throws if less than 0', () {
      expect(() => exactOut.minimumAmountOut(Percent(BigInt.from(-1), denominator: BigInt.from(100))),throwsA(isA<AssertionError>()));
    });

    test('returns exact if 0', () {
      expect(exactOut.minimumAmountOut(Percent(BigInt.from(0),denominator: BigInt.from(100))),exactOut.outputAmount);
    });

    test('returns slippage amount if nonzero', () {
      expect(exactOut.minimumAmountOut(Percent(BigInt.from(0),denominator: BigInt.from(100))), TokenAmount(token2, BigInt.from(100)));
      expect(exactOut.minimumAmountOut(Percent(BigInt.from(5),denominator: BigInt.from(100))),TokenAmount(token2, BigInt.from(100)));
      expect(exactOut.minimumAmountOut(Percent(BigInt.from(200),denominator: BigInt.from(100))), TokenAmount(token2, BigInt.from(100)));
    });
  });

  group('#bestTradeExactOut', () {
    // test('throws with empty pairs', () {
    //   expect(() => Trade.bestTradeExactOut([], token0, TokenAmount(token2, BigInt.from(100))),throwsA(isA<AssertionError>()));
    // });
    //
    // test('throws with max hops of 0', () {
    //   expect(() =>
    //       Trade.bestTradeExactOut([pair_0_2], token0, TokenAmount(token2,BigInt.from(100)), maxHops: 0 ),throwsA(isA<AssertionError>())
    //   );
    // });
    //
    // test('provides best route', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_2, pair_1_2],
    //       token0,
    //        TokenAmount(token2,BigInt.from(100))
    //   );
    //   expect(result.length,2);
    //   expect(result[0].route.pairs.length,1); // 0 -> 2 at 10:11
    //   expect(result[0].route.path, [token0, token2]);
    //   expect(result[0].inputAmount,TokenAmount(token0, BigInt.from(101)));
    //   expect(result[0].outputAmount, TokenAmount(token2, BigInt.from(100)) );
    //   expect(result[1].route.pairs.length, 2); // 0 -> 1 -> 2 at 12:12:10
    //   expect(result[1].route.path, [token0, token1, token2]);
    //   expect(result[1].inputAmount, TokenAmount(token0, BigInt.from(156)));
    //   expect(result[1].outputAmount, TokenAmount(token2, BigInt.from(100)));
    // });
    //
    // test('doesnt throw for zero liquidity pairs', (){
    //   expect(Trade.bestTradeExactOut([empty_pair_0_1], token1, TokenAmount(token1,BigInt.from(100))).length,0);
    // });
    //
    //
    // test('respects maxHops', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_2, pair_1_2],
    //       token0,
    //       TokenAmount(token2, BigInt.from(10)),
    //       maxHops: 1
    //   );
    //   expect(result.length,1);
    //   expect(result[0].route.pairs.length, 1); // 0 -> 2 at 10:11
    //   expect(result[0].route.path, [token0, token2]);
    // });
    //
    // test('insufficient liquidity', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_2, pair_1_2],
    //       token0,
    //       TokenAmount(token2,BigInt.from(1200))
    //   );
    //   expect(result.length, 0);
    // });
    //
    // test('insufficient liquidity in one pair but not the other', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_2, pair_1_2],
    //       token0,
    //       TokenAmount(token2, BigInt.from(1050))
    //   );
    //   expect(result.length,1);
    // });
    //
    // test('respects n', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_2, pair_1_2],
    //       token0,
    //       TokenAmount(token2, BigInt.from(10)),
    //     maxNumResults: 1
    // );
    //   expect(result.length,1);
    // });
    //
    // test('no path', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_0_1, pair_0_3, pair_1_3],
    //       token0,
    //       TokenAmount(token2, BigInt.from(10))
    //   );
    //   expect(result.length,0);
    // });

    // test('works for ETHER currency input', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_weth_0, pair_0_1, pair_0_3, pair_1_3],
    //       ETHER,
    //       TokenAmount(token3, BigInt.from(100)));
    //   expect(result.length, 2);
    //   // expect(result[0].inputAmount.currency, ETHER);
    //   // expect(result[0].route.path,
    //   //     [Token.WETH[ChainId.MAINNET.value], token0, token1, token3]);
    //   // expect(result[0].outputAmount.currency, token3);
    //   // expect(result[1].inputAmount.currency, ETHER);
    //   // expect(result[1].route.path,
    //   //     [Token.WETH[ChainId.MAINNET.value], token0, token3]);
    //   // expect(result[1].outputAmount.currency, token3);
    // });

    // test('works for ETHER currency output', () {
    //   final result = Trade.bestTradeExactOut(
    //       [pair_weth_0, pair_0_1, pair_0_3, pair_1_3],
    //       token3,
    //       CurrencyAmount.ether(BigInt.from(100)));
    //   expect(result.length, 2);
    //   expect(result[0].inputAmount.currency, token3);
    //   expect(result[0].route.path,
    //       [token3, token0, Token.WETH[ChainId.MAINNET.value]]);
    //   expect(result[0].outputAmount.currency, ETHER);
    //   expect(result[1].inputAmount.currency, token3);
    //   expect(result[1].route.path,
    //       [token3, token1, token0, Token.WETH[ChainId.MAINNET.value]]);
    //   expect(result[1].outputAmount.currency, ETHER);
    // });
  });
}
